@@ -15,7 +15,6 @@
 
 #include <stdio.h>
 #include <malloc.h>
-#include <unistd.h>
 #include <sys/param.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -45,7 +44,6 @@ struct lookup_context {
 int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */
 
 exports rpc_get_exports(const char *host, long seconds, long micros, unsigned int option);
-exports rpc_exports_prune(exports list);
 void rpc_exports_free(exports list);
 
 int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
@@ -58,7 +56,7 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	ctxt = malloc(sizeof(struct lookup_context));
 	if (!ctxt) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		crit(LOGOPT_ANY, MODPREFIX "malloc: %s", estr);
+		logerr(MODPREFIX "malloc: %s", estr);
 		return 1;
 	}
 
@@ -66,7 +64,7 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 
 	ctxt->parse = open_parse(mapfmt, MODPREFIX, argc, argv);
 	if (!ctxt->parse) {
-		crit(LOGOPT_ANY, MODPREFIX "failed to open parse context");
+		logerr(MODPREFIX "failed to open parse context");
 		free(ctxt);
 		return 1;
 	}
@@ -80,8 +78,160 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	return NSS_STATUS_UNKNOWN;
 }
 
+static char *get_exports(struct autofs_point *ap, const char *host)
+{
+	char buf[MAX_ERR_BUF];
+	char *mapent;
+	exports exp;
+
+	debug(ap->logopt, MODPREFIX "fetchng export list for %s", host);
+
+	exp = rpc_get_exports(host, 10, 0, RPC_CLOSE_NOLINGER);
+
+	mapent = NULL;
+	while (exp) {
+		if (mapent) {
+			int len = strlen(mapent) + 1;
+
+			len += strlen(host) + 2*(strlen(exp->ex_dir) + 2) + 3;
+			mapent = realloc(mapent, len);
+			if (!mapent) {
+				char *estr;
+				estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				error(ap->logopt, MODPREFIX "malloc: %s", estr);
+				rpc_exports_free(exp);
+				return NULL;
+			}
+			strcat(mapent, " \"");
+			strcat(mapent, exp->ex_dir);
+			strcat(mapent, "\"");
+		} else {
+			int len = 2*(strlen(exp->ex_dir) + 2) + strlen(host) + 3;
+
+			mapent = malloc(len);
+			if (!mapent) {
+				char *estr;
+				estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				error(ap->logopt, MODPREFIX "malloc: %s", estr);
+				rpc_exports_free(exp);
+				return NULL;
+			}
+			strcpy(mapent, "\"");
+			strcat(mapent, exp->ex_dir);
+			strcat(mapent, "\"");
+		}
+		strcat(mapent, " \"");
+		strcat(mapent, host);
+		strcat(mapent, ":");
+		strcat(mapent, exp->ex_dir);
+		strcat(mapent, "\"");
+
+		exp = exp->ex_next;
+	}
+	rpc_exports_free(exp);
+
+	if (!mapent)
+		error(ap->logopt, MODPREFIX "exports lookup failed for %s", host);
+
+	return mapent;
+}
+
+static int do_parse_mount(struct autofs_point *ap, struct map_source *source,
+			  const char *name, int name_len, char *mapent,
+			  struct lookup_context *ctxt)
+{
+	int ret;
+
+	master_source_current_wait(ap->entry);
+	ap->entry->current = source;
+
+	ret = ctxt->parse->parse_mount(ap, name, name_len,
+				 mapent, ctxt->parse->context);
+	if (ret) {
+		time_t now = time(NULL);
+		struct mapent_cache *mc = source->mc;
+		struct mapent *me;
+		int rv = CHE_OK;
+
+		cache_writelock(mc);
+		me = cache_lookup_distinct(mc, name);
+		if (me)
+			rv = cache_push_mapent(me, NULL);
+		else
+			rv = cache_update(mc, source, name, NULL, now);
+		if (rv != CHE_FAIL) {
+			me = cache_lookup_distinct(mc, name);
+			if (me)
+				me->status = now + ap->negative_timeout;
+		}
+		cache_unlock(mc);
+		return NSS_STATUS_TRYAGAIN;
+	}
+	return NSS_STATUS_SUCCESS;
+}
+
+static int update_hosts_mounts(struct autofs_point *ap,
+			       struct map_source *source, time_t age,
+			       struct lookup_context *ctxt)
+{
+	struct mapent_cache *mc;
+	struct mapent *me;
+	char *mapent;
+	int ret;
+
+	mc = source->mc;
+
+	pthread_cleanup_push(cache_lock_cleanup, mc);
+	cache_writelock(mc);
+	me = cache_lookup_first(mc);
+	while (me) {
+		/* Hosts map entry not yet expanded or already expired */
+		if (!me->multi)
+			goto next;
+
+		debug(ap->logopt, MODPREFIX "get list of exports for %s", me->key);
+
+		mapent = get_exports(ap, me->key);
+		if (mapent) {
+			cache_update(mc, source, me->key, mapent, age);
+			free(mapent);
+		}
+next:
+		me = cache_lookup_next(mc, me);
+	}
+	pthread_cleanup_pop(1);
+
+	pthread_cleanup_push(cache_lock_cleanup, mc);
+	cache_readlock(mc);
+	me = cache_lookup_first(mc);
+	while (me) {
+		/*
+		 * Hosts map entry not yet expanded, already expired
+		 * or not the base of the tree
+		 */
+		if (!me->multi || me->multi != me)
+			goto cont;
+
+		debug(ap->logopt, MODPREFIX
+		      "attempt to update exports for exports for %s", me->key);
+
+		master_source_current_wait(ap->entry);
+		ap->entry->current = source;
+		ap->flags |= MOUNT_FLAG_REMOUNT;
+		ret = ctxt->parse->parse_mount(ap, me->key, strlen(me->key),
+					       me->mapent, ctxt->parse->context);
+		ap->flags &= ~MOUNT_FLAG_REMOUNT;
+cont:
+		me = cache_lookup_next(mc, me);
+	}
+	pthread_cleanup_pop(1);
+
+	return NSS_STATUS_SUCCESS;
+}
+
 int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 {
+	struct lookup_context *ctxt = (struct lookup_context *) context;
 	struct map_source *source;
 	struct mapent_cache *mc;
 	struct hostent *host;
@@ -93,9 +243,24 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 
 	mc = source->mc;
 
+	debug(ap->logopt, MODPREFIX "read hosts map");
+
+	/*
+	 * If we don't need to create directories then there's no use
+	 * reading the map. We always need to read the whole map for
+	 * direct mounts in order to mount the triggers.
+	 */
+	if (!(ap->flags & MOUNT_FLAG_GHOST) && ap->type != LKP_DIRECT) {
+		debug(ap->logopt, MODPREFIX
+		      "map not browsable, update existing host entries only");
+		update_hosts_mounts(ap, source, age, ctxt);
+		source->age = age;
+		return NSS_STATUS_SUCCESS;
+	}
+
 	status = pthread_mutex_lock(&hostent_mutex);
 	if (status) {
-		error(LOGOPT_ANY, MODPREFIX "failed to lock hostent mutex");
+		error(ap->logopt, MODPREFIX "failed to lock hostent mutex");
 		return NSS_STATUS_UNAVAIL;
 	}
 
@@ -111,8 +276,9 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 
 	status = pthread_mutex_unlock(&hostent_mutex);
 	if (status)
-		error(LOGOPT_ANY, MODPREFIX "failed to unlock hostent mutex");
+		error(ap->logopt, MODPREFIX "failed to unlock hostent mutex");
 
+	update_hosts_mounts(ap, source, age, ctxt);
 	source->age = age;
 
 	return NSS_STATUS_SUCCESS;
@@ -124,11 +290,9 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	struct map_source *source;
 	struct mapent_cache *mc;
 	struct mapent *me;
-	char buf[MAX_ERR_BUF];
 	char *mapent = NULL;
 	int mapent_len;
 	time_t now = time(NULL);
-	exports exp;
 	int ret;
 
 	source = ap->entry->current;
@@ -137,6 +301,32 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 
 	mc = source->mc;
 
+	/* Check if we recorded a mount fail for this key anywhere */
+	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
+	if (me) {
+		if (me->status >= time(NULL)) {
+			cache_unlock(me->mc);
+			return NSS_STATUS_NOTFOUND;
+		} else {
+			struct mapent_cache *smc = me->mc;
+			struct mapent *sme;
+
+			if (me->mapent)
+				cache_unlock(smc);
+			else {
+				cache_unlock(smc);
+				cache_writelock(smc);
+				sme = cache_lookup_distinct(smc, name);
+				/* Negative timeout expired for non-existent entry. */
+				if (sme && !sme->mapent) {
+					if (cache_pop_mapent(sme) == CHE_FAIL)
+						cache_delete(smc, name);
+				}
+				cache_unlock(smc);
+			}
+		}
+	}
+
 	cache_readlock(mc);
 	me = cache_lookup_distinct(mc, name);
 	if (!me) {
@@ -145,7 +335,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		 * We haven't read the list of hosts into the
 		 * cache so go straight to the lookup.
 		 */
-		if (!ap->ghost) {
+		if (!(ap->flags & MOUNT_FLAG_GHOST)) {
 			/*
 			 * If name contains a '/' we're searching for an
 			 * offset that doesn't exist in the export list
@@ -158,10 +348,10 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		}
 
 		if (*name == '/')
-			msg(MODPREFIX
+			info(ap->logopt, MODPREFIX
 			      "can't find path in hosts map %s", name);
 		else
-			msg(MODPREFIX
+			info(ap->logopt, MODPREFIX
 			      "can't find path in hosts map %s/%s",
 			      ap->path, name);
 
@@ -177,104 +367,34 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	if (*name == '/') {
 		pthread_cleanup_push(cache_lock_cleanup, mc);
 		mapent_len = strlen(me->mapent);
-		mapent = alloca(mapent_len + 1);
+		mapent = malloc(mapent_len + 1);
 		if (mapent)
 			strcpy(mapent, me->mapent);
 		pthread_cleanup_pop(0);
 	}
 	cache_unlock(mc);
 
-	if (mapent) {
-		master_source_current_wait(ap->entry);
-		ap->entry->current = source;
-
-		debug(ap->logopt, MODPREFIX "%s -> %s", name, me->mapent);
-
-		ret = ctxt->parse->parse_mount(ap, name, name_len,
-				 mapent, ctxt->parse->context);
-
-		if (!ret)
-			return NSS_STATUS_SUCCESS;
-
-		return NSS_STATUS_TRYAGAIN;
-	}
 done:
-	/*
-	 * Otherwise we need to get the exports list and add update
-	 * the cache.
-	 */
-	debug(ap->logopt, MODPREFIX "fetchng export list for %s", name);
-
-	exp = rpc_get_exports(name, 10, 0, RPC_CLOSE_NOLINGER);
-
-	/* Check exports for obvious ones we don't have access to */
-	/*exp = rpc_exports_prune(exp);*/
-
-	mapent = NULL;
-	while (exp) {
-		if (mapent) {
-			int len = strlen(mapent) + 1;
-
-			len += strlen(name) + 2*strlen(exp->ex_dir) + 3;
-			mapent = realloc(mapent, len);
-			if (!mapent) {
-				char *estr;
-				estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				crit(ap->logopt, MODPREFIX "malloc: %s", estr);
-				rpc_exports_free(exp);
-				return NSS_STATUS_UNAVAIL;
-			}
-			strcat(mapent, " ");
-			strcat(mapent, exp->ex_dir);
-		} else {
-			int len = 2*strlen(exp->ex_dir) + strlen(name) + 3;
-
-			mapent = malloc(len);
-			if (!mapent) {
-				char *estr;
-				estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				crit(ap->logopt, MODPREFIX "malloc: %s", estr);
-				rpc_exports_free(exp);
-				return NSS_STATUS_UNAVAIL;
-			}
-			strcpy(mapent, exp->ex_dir);
-		}
-		strcat(mapent, " ");
-		strcat(mapent, name);
-		strcat(mapent, ":");
-		strcat(mapent, exp->ex_dir);
-
-		exp = exp->ex_next;
-	}
-	rpc_exports_free(exp);
-
-	/* Exports lookup failed so we're outa here */
-	if (!mapent) {
-		error(ap->logopt, "exports lookup failed for %s", name);
-		return NSS_STATUS_UNAVAIL;
-	}
-
 	debug(ap->logopt, MODPREFIX "%s -> %s", name, mapent);
 
-	cache_writelock(mc);
-	cache_update(mc, source, name, mapent, now);
-	cache_unlock(mc);
+	if (!mapent) {
+		/* We need to get the exports list and update the cache. */
+		mapent = get_exports(ap, name);
 
-	debug(LOGOPT_ANY, "source wait");
+		/* Exports lookup failed so we're outa here */
+		if (!mapent)
+			return NSS_STATUS_UNAVAIL;
 
-	master_source_current_wait(ap->entry);
-	ap->entry->current = source;
+		cache_writelock(mc);
+		cache_update(mc, source, name, mapent, now);
+		cache_unlock(mc);
+	}
 
-	debug(LOGOPT_ANY, "do parse_mount");
+	ret = do_parse_mount(ap, source, name, name_len, mapent, ctxt);
 
-	ret = ctxt->parse->parse_mount(ap, name, name_len,
-				 mapent, ctxt->parse->context);
 	free(mapent);
 
-	if (ret)
-		return NSS_STATUS_TRYAGAIN;
-
-	return NSS_STATUS_SUCCESS;
+	return ret;
 }
 
 int lookup_done(void *context)
